from __future__ import print_function
import sys
import time
import functools

import zope.interface
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.endpoints import serverFromString
from twisted.web.http import HTTPChannel
from twisted.web.static import Data
from twisted.web.resource import Resource
from twisted.web.server import Site

from carml.interface import ICarmlCommand
from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


class TempHSOptions(usage.Options):
    """
    """

    optFlags = [
    ]

    optParameters = [
    ]

    def __init__(self):
        usage.Options.__init__(self)
        self['ports'] = []

    def opt_port(self, port):
        "add a port redirect: with a colon, specify a different local port (e.g. \"80\" to pass through or \"80:9876\" to listen locally on 9876). Can be used multiple times."
        self['ports'].append(port)


class TempHSCommand(object):
    """
    Keep an ephemeral hidden-service active as long as this command is
    running.
    """
    zope.interface.implements(ICarmlCommand, IPlugin)

    name = 'temphs'
    help_text = 'Add a temporary hidden-service to the Tor we connect to.'
    build_state = False
    controller_connection = True
    options_class = TempHSOptions

    def validate(self, options, mainoptions):
        "ICarmlCommand API"
        if len(options['ports']) == 0:
            raise RuntimeError("Must specify at least one --port")

    def error(self, e):
        print("ERR: %s" % e)

    @inlineCallbacks
    def run(self, options, mainoptions, proto):
        "ICarmlCommand API"

        def info(msg):
            if 'Service descriptor (v2) stored' in msg:
                got_upload.callback(None)
        got_upload = Deferred()
        proto.add_event_listener('INFO', info)

        hs = txtorcon.EphemeralHiddenService(options['ports'])
        yield hs.add_to_tor(proto)
        print("Created HS", hs.hostname)

        def remove():
            print("removing hidden-service")
            return hs.remove_from_tor(proto)
        reactor.addSystemEventTrigger('before', 'shutdown', remove)

        print("...waiting for descriptor upload")
        yield got_upload
        print("Got one.", time.asctime())

        # we never callback() on this, so we serve forever
        d = Deferred()
        yield d

    def _remove_service(self, config, hs):
        print("removing " + hs)
        config.hiddenservices.remove(hs)
        return config.save()

# the IPlugin/getPlugin stuff from Twisted picks up any object from
# here than implements ICarmlCommand -- so we need to instantiate one
cmd = TempHSCommand()



from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log
from twisted.protocols.basic import LineReceiver
from twisted.internet import defer, stdio

from PyQt4 import QtGui

import txtorcon
from carml.interface import ICarmlCommand
from carml import util


class ConfigDruidOptions(usage.Options):
    def __init__(self):
        super(ConfigDruidOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self.args = None


def do_cmd(proto, args):
    def _print(res):
        print(res)

    def _error(arg):
        print(util.colors.red(arg.getErrorMessage()))
        return None
    d = defer.Deferred()
    # launch qt gui? in own thread, or what? how do we integrate here?
    # Create a Qt application

    app = QtGui.QApplication(sys.argv)
    # Create a Label and show it
    label = QtGui.QLabel("Hello World")
    label.show()
    # Enter Qt application main loop
#    app.exec_()
#    sys.exit()

    return d


# see cmd_info for an alternate way to implement this via a method
# with attributes and "zope.interface.implementsDirectly()"
# trying out both ways to see what feels better
@zope.interface.implementer(ICarmlCommand)
class ConfigDruidCommand(object):
    # Attributes specified by ICarmlCommand
    name = 'configdruid'
    help_text = 'Launch a Qt-based interactive configuration druid.'
    controller_connection = True
    build_state = False
    options_class = ConfigDruidOptions

    def validate(self, options, mainoptions):
        pass

    @defer.inlineCallbacks
    def run(self, options, mainoptions, proto):
        """
        ICarmlCommand API
        """

        yield do_cmd(proto, options.args)

cmd = ConfigDruidCommand()
__all__ = ['cmd']



from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log
from twisted.protocols.basic import LineReceiver
from twisted.internet import defer, reactor, stdio

import txtorcon
from carml.interface import ICarmlCommand
from carml import util


class HelpOptions(usage.Options):
    def __init__(self):
        super(HelpOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self.what = None

    def parseArgs(self, *args):
        self.what = args[0] if len(args) else None

    def getUsage(self, **kw):
        return "Options:\n   Pass a single sub-command to print help for."


@zope.interface.implementer(ICarmlCommand)
class HelpSubCommand(object):
    # Attributes specified by ICarmlCommand
    name = 'help'
    help_text = 'Print help on sub-commands (like "carml help events").'
    controller_connection = False
    build_state = False
    options_class = HelpOptions

    def validate(self, options, mainoptions):
        pass

    def run(self, options, mainoptions, proto):
        """
        ICarmlCommand API
        """

        w = options.what
        if w is None:
            from carml.dispatch import Options
            print(Options().getUsage())
            return
            for sub in mainoptions.commands.values():
                print('## %s\n%s\n' % (sub.name, ('-' * (3 + len(sub.name)))))
                print_help_for(sub)
            return

        try:
            print('Sub-command "%s":' % w)
            print_help_for(mainoptions.commands[w])
        except KeyError:
            print('Unknown command "%s".' % w)
        return defer.succeed(None)


def print_help_for(sub):
    desc = util.wrap(sub.help_text, 60, '    ')
    print(desc)
    for line in sub.options_class().getUsage().split('\n'):
        print('    ', line)


cmd = HelpSubCommand()
__all__ = ['cmd']



from __future__ import print_function

import sys
import datetime
import functools
import random

from twisted.python import usage, log
from twisted.internet import defer, reactor, endpoints
import zope.interface
import humanize
import txtorcon

from carml.interface import ICarmlCommand
from carml import util


class RelayOptions(usage.Options):
    def __init__(self):
        super(RelayOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self['delete'] = []

    optFlags = [
        ('list', None, 'List all relays, by hex ID.'),
    ]

    optParameters = [
        ('info', None, '', 'Look up by fingerprint (or part of one).', str),
        ('await', None, '', 'Monitor NEWCONSENSUS for a fingerprint to exist', str),
    ]


@defer.inlineCallbacks
def _print_router_info(router, agent=None):
    #loc = yield router.get_location()
    loc = yield router.location
    print("            name: {}".format(router.name))
    print("          hex id: {}".format(router.id_hex))
    print("        location: {}".format(loc.countrycode))
    print("         address: {}:{} (DirPort={})".format(router.ip, router.or_port, router.dir_port))
    diff = datetime.datetime.utcnow() - router.modified
    print("  last published: {} ago".format(humanize.naturaldelta(diff)))
    if agent:
        print(util.colors.italic("Extended information from" + util.colors.green(" onionoo.torproject.org") + ":"))
        details = yield router.get_onionoo_details(agent)
        details['or_addresses'] = ', '.join(details['or_addresses'])
        if False:
            details = {
                k: util.colors.bold(str(v))
                for k, v in details.items()
            }
        print(
            "        platform: {platform}\n"
            "        runnning: {running}\n"
            "     dir_address: {dir_address}\n"
            "    OR addresses: {or_addresses}\n"
            "        location: {city_name}, {region_name}, {country_name}\n"
            "       host name: {host_name}\n"
            "              AS: {as_number} ({as_name})\n"
            "  last restarted: {last_restarted}\n"
            "    last changed: {last_changed_address_or_port}\n"
            "       last seen: {last_seen}\n"
            "   probabilities: guard={guard_probability} middle={middle_probability} exit={exit_probability}\n"
            "".format(**details)
        )


@defer.inlineCallbacks
def router_info(state, arg, tor):
    if len(arg) == 40 and not arg.startswith('$'):
        arg = '${}'.format(arg)
    try:
        yield _print_router_info(state.routers[arg], agent=tor.web_agent())
    except KeyError:
        candidates = [
            r for r in state.all_routers
            if arg in r.name or arg in r.id_hex
        ]
        if not candidates:
            print("Nothing found ({} routers total)".format(len(state.all_routers)))
        if len(candidates) > 1:
            print("Found multiple routers:")
        for router in candidates:
            yield _print_router_info(router)
            print()


def _when_updated(state):
    d = defer.Deferred()
    def _newconsensus(doc):
        # we actually don't care what's *in* the event, we just know
        # that the state has now updated...maybe a .when_updated() in
        # TorState?
        print("Got NEWCONSENSUS at {}".format(datetime.datetime.now()))
        d.callback(None)
        state.protocol.remove_event_listener('NEWCONSENSUS', _newconsensus)
    state.protocol.add_event_listener('NEWCONSENSUS', _newconsensus)
    return d


@defer.inlineCallbacks
def _await_router(state, router_id):
    while True:
        yield _when_updated(state)
        try:
            defer.returnValue(state.routers[router_id])
            return
        except KeyError:
            continue


@defer.inlineCallbacks
def router_await(state, arg):
    if len(arg) == 40 and not arg.startswith('$'):
        arg = '${}'.format(arg)
    if not arg.startswith('$') and len(arg) == 41:
        print("Doesn't appear to be a hex router ID")
        return

    if arg in state.routers:
        print("Router already present:")
        r = state.routers[arg]
    else:
        r = yield _await_router(state, arg)
    yield _print_router_info(r)


def router_list(state):
    for router in state.all_routers:
        print("{}".format(router.id_hex[1:]))


class RelayCommand(object):
    zope.interface.implements(ICarmlCommand)

    # Attributes specified by ICarmlCommand
    name = 'relay'
    options_class = RelayOptions
    help_text = 'Information about Tor relays.'
    controller_connection = True
    build_state = True

    def validate(self, options, mainoptions):
        """ICarmlCommand API"""

    @defer.inlineCallbacks
    def run(self, options, mainoptions, state):
        """
        ICarmlCommand API
        """
        endpoint_str = mainoptions['connect']
        endpoint = endpoints.clientFromString(reactor, endpoint_str)
        tor = yield txtorcon.connect(reactor, endpoint)
        if options['info']:
            yield router_info(state, options['info'], tor)
        elif options['list']:
            yield router_list(state)
        elif options['await']:
            yield router_await(state, options['await'])



cmd = RelayCommand()
__all__ = ['cmd']



# This is intended to be copy-pasted, and search-replace "blam" with
# your command-name as a starting-point for a new command. To use:
#
# 1. copy/paste to `cmdname.py` in a directory and with PYTHONPATH
#    such that it's in a carml/command sub-directory. E.g. adding
#    `example_extension` from the source code to PYTHONPATH will make the
#    `blam` command avaliable:
#
#        virtualenv venv
#        . ./venv/bin/acvtivate
#        python setup.py develop
#        carml blam --amaze 5
#
# 2. delete the example optFlags and optParameters entries and add any
#    of your own.
#
#        http://twistedmatrix.com/documents/current/core/howto/options.html
#
# 3. change `name` and `help_text` members of the command object
#
# 4. Decide if you need no Tor connection (e.g. you're going to launch
#    your own, like "pastebin" command), or if you need a TorState
#    object instead (e.g. you want to use the high-level circuit or
#    stream or event interface) and if so, whether to load all the
#    routers (relays) or not. Then set build_state, load_routers and
#    controller_connection as appropriate.
#
# 5. gut validate() method and do any options-validation you need
#    to. Note you have no Tor connection at this point; if you require
#    one to do validation, you have to error out of run() instead.
#
# 6. implement run() method. This may return:
#
#    a. A Deferred instance, in which case carml command-line exits
#       when the Deferred callback()s or errback()s.
#    b. None, in which case the command itself must call
#       reactor.stop() exactly once in order to exit
#
#    In either case, you can easily make a command that never
#    exits. (It will clean up if user does control-c or sends TERM
#    etc.)

import zope.interface
from twisted.python import usage
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet import defer

from carml.interface import ICarmlCommand

class BlamOptions(usage.Options):
    """
    See the Twisted docs for option arguments, but some common
    examples are below.
    """

    optFlags = [
        ('booooring', 'b', 'If specified, be boring.'),
        ]

    optParameters = [
        ('amaze', 'a', 0, 'Specify an amazement level. 0-10', int)
        ]


@zope.interface.implementer(ICarmlCommand)
@zope.interface.implementer(IPlugin)
class BlamCommand(object):
    """
    The actual command, which implements ICarmlCommand and Twisted's
    IPlugin.
    
    note to self (FIXME): can I just have ICarmlCommand derive from
    IPlugin?
    """

    # these are all Attributes of the ICarmlCommand interface
    name = 'blam'
    help_text = """A template command."""
    build_state = False
    load_routers = False
    controller_connection = True
    options_class = BlamOptions

    def validate(self, options, mainoptions):
        "ICarmlCommand API"

        if options['booooring'] and options['amaze']:
            raise RuntimeError("Can't specify both boredom *and* amazement.")
        if options['amaze'] < 0 or options['amaze'] >= 10:
            raise RuntimeError("Only amazing from 0-10...")
        return

    def run(self, options, mainoptions, connection):
        "ICarmlCommand API"

        if not options['booooring']:
            print "Blam!"
            if options['amaze']:
                for x in xrange(options['amaze']):
                    print ' ' + ('-' * x * 2) + '-> Such amaze!'
        return defer.succeed(self)

## the IPlugin/getPlugin stuff from Twisted picks up any object from
## this file than implements ICarmlCommand -- so we need to
## instantiate one
cmd = BlamCommand()



from __future__ import print_function

import os
import sys
import time
import functools

import zope.interface
from twisted.python import usage, log
from twisted.internet import defer, reactor

import txtorcon

from carml.interface import ICarmlCommand
from carml.util import dump_circuits
from carml.util import format_net_location
from carml.util import nice_router_name
from carml.util import colors


class EventsOptions(usage.Options):
    optFlags = [
        ('list', 'L', 'Show available events.'),
        ('once', '', 'Output exactly one and quit (same as -n 1 or --count=1).'),
        ('show-event', 's', 'Prefix each line with the event it is from.'),
    ]

    optParameters = [
        ('count', 'n', None, 'Output this many events, and quit (default is unlimited).', int),
    ]

    def parseArgs(self, *args):
        self['events'] = args

    def __init__(self):
        """
        We override this to get rid of the Twisted default --version and --help things
        """
        super(EventsOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self['log-level'] = []


class EventsCommand(object):
    zope.interface.implements(ICarmlCommand)

    name = 'events'
    help_text = """Follow any Tor events, listed as positional arguments."""
    controller_connection = True
    build_state = False
    options_class = EventsOptions

    def validate(self, options, mainoptions):
        if not options['list']:
            if not len(options['events']):
                raise RuntimeError("Need at least one event to watch.")

    def _got_event(self, evt, msg):
        if self.counter is not None:
            self.counter -= 1
            if self.counter >= 0:
                print(msg)
            elif self.all_done:
                self.all_done.callback(None)
                self.all_done = None
        elif evt:
            print("{}: {}".format(evt, msg))
        else:
            ts = time.asctime()
            print("{} {}".format(ts, msg))

    @defer.inlineCallbacks
    def run(self, options, mainoptions, proto):
        all_events = yield proto.get_info('events/names')
        if options['list']:
            for e in all_events['events/names'].split():
                print(e)
            return

        self.all_done = defer.Deferred()
        self.counter = options['count']
        if options['once']:
            self.counter = 1

        for e in options['events']:
            e = e.upper()
            if e not in all_events['events/names']:
                print("Invalid event:", e)
                return
            if options['show-event']:
                listener = functools.partial(self._got_event, e)
            else:
                listener = functools.partial(self._got_event, None)
            proto.add_event_listener(e, listener)

        # might be forever if there's no count
        yield self.all_done


cmd = EventsCommand()
__all__ = ['cmd']



from __future__ import print_function
import sys
import time
import functools

from zope.interface import implementer
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.protocol import Protocol


from carml.interface import ICarmlCommand
from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


class TmuxOptions(usage.Options):
    """
    """

    optFlags = [
    ]

    optParameters = [
    ]


@implementer(ICarmlCommand, IPlugin)
class TmuxCommand(object):
    """
    Output for tmux's status-right. As in, put this in ~/.tmux.conf::

        set-option -g status-utf8 on
        set-option -g status-fg green
        set -g status-right '#(carml tmux)'
        set -g status-interval 2
    """

    name = 'tmux'
    help_text = """Show some informantion in your tmux status."""
    build_state = True
    load_routers = False
    controller_connection = True
    options_class = TmuxOptions

    def validate(self, options, mainoptions):
        "ICarmlCommand API"
        pass

    @defer.inlineCallbacks
    def run(self, options, mainoptions, state):
        "ICarmlCommand API"

        # because we asked not to load routers (for speed) we do have
        # to "update" the one's we'll be interested in.
        if False:
            routers = set()
            for circ in state.circuits.values():
                for r in circ.path:
                    routers.add(r.id_hex)
            yield state.update_routers(routers)

        for circ in state.circuits.values():
            if len(circ.streams) == 0:
                continue

            for r in circ.path:
                if r.location.countrycode is None:
                    yield r.get_country()
            # print(circ)
            path = u'>'.join(map(lambda r: r.location.countrycode or '__', circ.path))
            #print(u"#[fg=colour46,bg=colour28]{}#[fg=colour28,bg=colour22]{}".format(path, len(circ.streams)), end='')
            print(u"#[fg=colour28,bg=colour22]{}#[fg=colour46,bg=colour28]{}".format(path, len(circ.streams)), end='')

        total = len(state.circuits)
        general = len(filter(lambda c: c.purpose == 'GENERAL', state.circuits.values()))
        part = int(general / float(total) * 7)
        msg = u'#[fg=colour32,bg=colour17]%s' % unichr(0x2581 + part)
        print(msg.encode('utf8'), end='')

        onions = len(filter(lambda c: c.purpose.startswith('HS_'), state.circuits.values()))
        # print("XXX", general, onions)
        part = int(onions / float(total) * 7)
        msg = u'#[fg=colour160,bg=colour17]%s' % unichr(0x2581 + part)
        print(msg.encode('utf8'))
        # util.dump_circuits(state, True)
        # stuff like: #[fg=colour46,bg=colour28]

# the IPlugin/getPlugin stuff from Twisted picks up any object from
# here than implements ICarmlCommand -- so we need to instantiate one
cmd = TmuxCommand()



from __future__ import print_function

import sys
import datetime
import functools
import random

from twisted.python import usage, log
from twisted.internet import defer, reactor
import zope.interface
import humanize
import txtorcon

from carml.interface import ICarmlCommand
from carml import util


class CircOptions(usage.Options):
    def __init__(self):
        super(CircOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self['delete'] = []

    optFlags = [
        ('list', 'L', 'List existing circuits.'),
        ('if-unused', 'u', 'When deleting, pass the IfUnused flag to Tor.'),
        ('verbose', None, 'More information per circuit.'),
    ]

    optParameters = [
        ('build', 'b', '', 'Build a new circuit, given a comma-separated list of router names or IDs. Use "auto" to let Tor select the route.', str),
    ]

    def opt_delete(self, arg):
        """Delete a circuit by its ID."""
        for x in arg.split(','):
            self['delete'].append(int(x))


@defer.inlineCallbacks
def list_circuits(options, proto):
    print("Circuits:")
    state = txtorcon.TorState(proto)
    yield state.post_bootstrap

    now = datetime.datetime.utcnow()
    util.dump_circuits(state, options['verbose'])


@defer.inlineCallbacks
def delete_circuit(proto, circid, ifunused):
    # the already_deleted voodoo is because sometimes the circuit is
    # already marked as deleted before the OK comes back from the
    # controller, as in you get the event "first".
    # perhaps txtorcon should "fix"/normalize that such that it saves
    # events until the OK? maybe tor bug?
    class DetermineCircuitClosure(object):
        def __init__(self, target_id, done_deferred):
            self.circ_id = str(target_id)
            self.circuit_gone = False
            self.already_deleted = False
            self.completed = done_deferred

        def __call__(self, text):
            cid, what, _ = text.split(' ', 2)
            if what in ['CLOSED', 'FAILED']:
                if self.circ_id == cid:
                    self.circuit_gone = True
                    print("...circuit %s gone." % self.circ_id)
                    sys.stdout.flush()
                    if not self.already_deleted:
                        self.completed.callback(self)

    unused_string = '(if unused) ' if ifunused else ''
    print('Deleting circuit %s"%s"...' % (unused_string, circid),)

    gone_d = defer.Deferred()
    monitor = DetermineCircuitClosure(circid, gone_d)
    proto.add_event_listener('CIRC', monitor)
    sys.stdout.flush()

    state = txtorcon.TorState(proto, bootstrap=False)
    kw = {}
    if ifunused:
        kw['IfUnused'] = True
    try:
        status = yield state.close_circuit(circid, **kw)
        monitor.already_deleted = True
    except txtorcon.TorProtocolError as e:
        gone_d.errback(e)
        yield gone_d
        return

    if monitor.circuit_gone:
        return

    print(status, '(waiting for CLOSED)...')
    sys.stdout.flush()
    yield gone_d
    # we're now awaiting a callback via CIRC events indicating
    # that our circuit has entered state CLOSED


class _BuiltCircuitListener(txtorcon.CircuitListenerMixin):
    def __init__(self, circid, all_done):
        self.circid = circid
        self.first = True
        self._all_done = all_done

    def circuit_extend(self, circuit, router):
        if circuit.id == self.circid:
            if self.first:
                self.first = False
            else:
                sys.stdout.write('->')
            sys.stdout.write(router.name)
            sys.stdout.flush()

    def circuit_built(self, circuit):
        if circuit.id == self.circid:
            print(": " + util.colors.green("built."))
            self._all_done.callback(None)

    def circuit_closed(self, circuit, **kw):
        if circuit.id == self.circid:
            print(": " + util.colors.red("closed."))
            self._all_done.callback(None)

    def circuit_failed(self, circuit, **kw):
        if circuit.id == self.circid:
            r = kw['reason'] if 'reason' in kw else ''
            rr = kw['remote_reason'] if 'remote_reason' in kw else ''
            msg = util.colors.red('failed') + ' (%s, %s).' % (r, rr)
            self._all_done.errback(RuntimeError(msg))


@defer.inlineCallbacks
def build_circuit(proto, routers):
    state = txtorcon.TorState(proto)
    yield state.post_bootstrap
    if len(routers) == 1 and routers[0].lower() == 'auto':
        routers = None
        # print("Building new circuit, letting Tor select the path.")
    else:
        def find_router(args):
            position, name = args
            if name == '*':
                if position == 0:
                    return random.choice(state.entry_guards.values())
                else:
                    return random.choice(state.routers.values())
            r = state.routers.get(name) or state.routers.get('$'+name)
            if r is None:
                if len(name) == 40:
                    print("Couldn't look up %s, but it looks like an ID" % name)
                    r = name
                else:
                    raise RuntimeError('Couldn\'t find router "%s".' % name)
            return r
        routers = map(find_router, enumerate(routers))
        print("Building circuit:", '->'.join(map(util.nice_router_name, routers)))

    try:
        circ = yield state.build_circuit(routers)
        all_done = defer.Deferred()

        sys.stdout.write("Circuit ID %d: " % circ.id)
        sys.stdout.flush()
        state.add_circuit_listener(_BuiltCircuitListener(circ.id, all_done))
        # all_done will callback when the circuit is built (or errback
        # if it fails).

    except txtorcon.TorProtocolError as e:
        log.err(e)

    yield all_done


class CircCommand(object):
    zope.interface.implements(ICarmlCommand)

    # Attributes specified by ICarmlCommand
    name = 'circ'
    options_class = CircOptions
    help_text = 'Manipulate Tor circuits.'
    controller_connection = True
    build_state = False

    def validate(self, options, mainoptions):
        """ICarmlCommand API"""
        if options['list'] == 0 and options['delete'] == [] and not options['build']:
            raise RuntimeError("Must specify one of --list, --delete, --build")
        if options['if-unused'] and options['delete'] == []:
            raise RuntimeError("--if-unused is only for use with --delete")

    def run(self, options, mainoptions, proto):
        """
        ICarmlCommand API
        """

        if options['list']:
            return list_circuits(options, proto)

        elif len(options['delete']) > 0:
            deletes = []
            for d in options['delete']:
                deletes.append(delete_circuit(proto, d, options['if-unused']))
            return defer.DeferredList(deletes)

        elif options['build']:
            return build_circuit(proto, options['build'].split(','))


cmd = CircCommand()
__all__ = ['cmd']



# check Python Package Index (PyPI) over 3 different circuits and
# compare sha256 hashes. suitable for use in requirements.txt with
# "twine"

from zope.interface import implementer
from twisted.python import usage
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet import endpoints
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import deferLater
from twisted.web.client import readBody

from carml.interface import ICarmlCommand

import json
import hashlib
from base64 import urlsafe_b64encode
from distutils.version import LooseVersion

import txtorcon
import txsocksx
import txsocksx.http
import txsocksx.errors


class CheckPyPIOptions(usage.Options):
    """
    See the Twisted docs for option arguments, but some common
    examples are below.
    """

    optFlags = [
    ]

    optParameters = [
        ('package', 'p', None, 'Name of the package to check (unfortunately, case matters)', str),
        ('revision', 'r', None, 'Specific version to check (default: latest)', str),
    ]


@implementer(ICarmlCommand)
@implementer(IPlugin)
class CheckPyPICommand(object):
    """
    The actual command, which implements ICarmlCommand and Twisted's
    IPlugin.

    note to self (FIXME): can I just have ICarmlCommand derive from
    IPlugin?
    """

    # these are all Attributes of the ICarmlCommand interface
    name = 'checkpypi'
    help_text = """Check a PyPI package hash across multiple circuits."""
    build_state = True
    load_routers = True
    controller_connection = True
    options_class = CheckPyPIOptions

    def validate(self, options, mainoptions):
        """ICarmlCommand API"""
        if not options['package']:
            raise RuntimeError("Must supply --package")
        return None

    @inlineCallbacks
    def run(self, options, mainoptions, state):
        """ICarmlCommand API"""

        package_name = options['package']
        package_version = options['revision']

        @implementer(txtorcon.IStreamAttacher)
        class MyAttacher(object):
            circ = None
            def attach_stream(self, stream, circuits):
                print "Via:", ' '.join(map(lambda r: r.id_hex, self.circ.path))
                return self.circ

        # ask Tor what port it's running on
        port = yield state.protocol.get_info('net/listeners/socks')
        socks = 'tcp:' + port['net/listeners/socks']
        tor_ep = endpoints.clientFromString(reactor, socks)
        agent = txsocksx.http.SOCKS5Agent(reactor, proxyEndpoint=tor_ep)

        # download metadata from PyPI over "any" circuit
        uri = 'https://pypi.python.org/pypi/%s/json' % package_name
        print "downloading:", uri
        resp = yield agent.request('GET', uri)
        data = yield readBody(resp)
        data = json.loads(data)

        # did we get a valid sdist URL somewhere?
        sdist_url = None
        version = None
        if package_version is None:
            # print data['releases'].keys()
            available = [LooseVersion(x) for x in data['releases'].keys()]
            package_version = str(sorted(available)[-1])
            print "Using version:", package_version

        for url in data['releases'][package_version]:
            if url['packagetype'] == 'sdist':
                sdist_url = url['url'].encode('UTF8')
                version = url['filename']
        if sdist_url is None:
            print "Error: couldn't find any 'sdist' URLs"
            raise RuntimeError("No sdist URL")
        else:
            print "Found sdist:", sdist_url, "for", version

        # create our custom attacher
        attach = MyAttacher()
        yield state.set_attacher(attach, reactor)

        # download the distribution over several different circuits,
        # and record the sha256 hash each time.
        digests = []
        while len(digests) < 3:
            circ = yield state.build_circuit()
            try:
                yield circ.when_built()
            except Exception:
                print "Circuit failed; trying another."
                continue
            attach.circ = circ
            print "Built circuit"

            try:
                resp = yield agent.request('GET', sdist_url)
                # FIXME could stream this to the hasher with a custom
                # protocol, but teh RAMz they are cheap
                tarball = yield readBody(resp)
            except txsocksx.errors.TTLExpired as e:
                print "Timed out: {}".format(e)
                continue
            except Exception as e:
                print "Something went wrong: {}".format(e)
                continue

            hasher = hashlib.new('sha256')
            hasher.update(tarball)
            # the whole point is to match peep's hashes, and this is
            # exactly what it does:
            digest = urlsafe_b64encode(hasher.digest()).decode('ascii').rstrip('=')
            digests.append((sdist_url, circ, digest))
            print "sha256:", digest

        print "Found hashes:"
        feel_fear = False
        for (url, circ, digest) in digests:
            print digest
            if digest != digests[0][-1]:
                print "Fearsome Warning! Mismatched digest!!"
                feel_fear = True
                print "Circuit:", '->'.join(map(lambda r: r.hex_id, circ.path))

        if feel_fear:
            print "****\n  Something fishy!\n****"
        else:
            print "Add this to requirements.txt for peep:"
            print
            print "# sha256: {}".format(digests[0][-1])
            print "{}=={}".format(package_name, package_version)
        return

cmd = CheckPyPICommand()



from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log, failure
from twisted.internet import defer, reactor, error
import humanize

import txtorcon

from twisted.internet.defer import inlineCallbacks, Deferred

from carml.interface import ICarmlCommand
from carml.util import dump_circuits, format_net_location, nice_router_name, colors, wrap

LOG_LEVELS = ["DEBUG", "INFO", "NOTICE", "WARN", "ERR"]


class GraphOptions(usage.Options):
    optFlags = [
    ]

    optParameters = [
        ('max', 'm', 1024*20, 'Maximum scale, in bytes.', int)
    ]

    def __init__(self):
        """
        We override this to get rid of the Twisted default --version and --help things
        """
        super(GraphOptions, self).__init__()
        self.longOpt.remove('version')
        ##self.longOpt.remove('help')

class BandwidthTracker(object):
    '''
    This tracks bandwidth usage.
    '''

    def __init__(self, maxscale, state):
        #: a list of tuples
        self._bandwidth = []
        self._max = float(maxscale)
        self._state = state

    def circuits(self):
        return len(self._state.circuits)

    def streams(self):
        return len(self._state.streams)

    def on_bandwidth(self, s):
        r, w = map(int, s.split())
        self._bandwidth.append((r, w))
        self.draw_bars()

    def on_stream_bandwidth(self, s):
        #print("stream", s)
        pass

    def draw_bars(self):
        up = min(1.0, self._bandwidth[-1][0] / self._max)
        dn = min(1.0, self._bandwidth[-1][1] / self._max)
        kbup = self._bandwidth[-1][1] / 1024.0
        kbdn = self._bandwidth[-1][0] / 1024.0

        status = ' ' + colors.green('%.2f' % kbdn)
        status += '/'
        status += colors.red('%.2f' % kbup) # + ' KiB write'
        status += ' KiB/s'
        status += ' (%d streams, %d circuits)' % (self.streams(), self.circuits())

        # include the paths of any currently-active streams
        streams = ''
        for stream in self._state.streams.values():
            # ...there's a window during which it may not be attached yet
            if stream.circuit:
                circpath = '>'.join(map(lambda r: r.location.countrycode, stream.circuit.path))
                streams += ' ' + circpath
        if len(streams) > 24:
            streams = streams[:21] + '...'
        print(left_bar(up, 20) + unichr(0x21f5) + right_bar(dn, 20) + status + streams)



def left_bar(percent, width):
    '''
    Creates the green/left bar from a percentage and width. It uses
    some unicode vertical-bar characters to get some extra precision
    on the bar-length.
    '''
    blocks = int(percent * width)
    remain = (percent * width) - blocks

    part = int(remain * 8)
    rpart = unichr(0x258f - 7 + part) # for smooth bar

    return (' ' * (width - blocks)) + colors.negative(colors.green(rpart)) + colors.green(('+' * (blocks)), bg='green')

def right_bar(percent, width):
    '''
    See left_bar(); inverse and in red instead.
    '''
    blocks = int(percent * width)
    remain = (percent * width) - blocks

    part = int(remain * 8)
    rpart = unichr(0x258f - part) # for smooth bar
    if part == 0:
        rpart = ' '

    return colors.red('+' * (blocks), bg='red') + (colors.red(rpart)) + (' ' * (width - blocks))

@inlineCallbacks
def main(options, state):
    proto = state.protocol
    bwtracker = BandwidthTracker(options['max'], state)
    yield proto.add_event_listener('BW', bwtracker.on_bandwidth)
    yield proto.add_event_listener('STREAM_BW', bwtracker.on_stream_bandwidth)

    # infinite loop
    yield Deferred()


class GraphCommand(object):
    zope.interface.implements(ICarmlCommand)

    name = 'graph'
    help_text = 'A nice coloured console bandwidth-graph.'
    controller_connection = True
    build_state = True
    options_class = GraphOptions

    def validate(self, options, mainoptions):
        pass

    def run(self, options, mainoptions, state):
        return main(options, state)


cmd = GraphCommand()
__all__ = ['cmd']



from __future__ import print_function

import sys
import functools

import zope.interface
from twisted.python import usage, log
from twisted.internet import defer, reactor

from carml.interface import ICarmlCommand


class NewidOptions(usage.Options):
    def __init__(self):
        super(NewidOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')


def newid_got_signal(all_done, x):
    if x == 'NEWNYM':
        print('success.')
        all_done.callback(None)


def newid_no_signal(left, all_done):
    if left > 0:
        print('Waiting %d more seconds.' % left)
        reactor.callLater(1, newid_no_signal, left - 1, all_done)
    else:
        all_done.errback(RuntimeError('no acknowledgement in 10 seconds.'))


@zope.interface.implementer(ICarmlCommand)
class NewidCommand(object):
    # Attributes specified by ICarmlCommand
    name = 'newid'
    help_text = 'Ask Tor for a new identity via NEWNYM, and listen for the response acknowledgement.'
    controller_connection = True
    build_state = False
    options_class = NewidOptions

    def validate(self, options, mainoptions):
        return

    @defer.inlineCallbacks
    def run(self, options, mainoptions, proto):
        all_done = defer.Deferred()

        # Tor emits this event whenever it processes a SIGNAL command.
        proto.add_event_listener('SIGNAL', functools.partial(newid_got_signal, all_done))

        print("Requesting new identity",)
        sys.stdout.flush()
        yield proto.signal('NEWNYM')
        # answer will be "OK" since Tor received the signal
        # if rate-limiting is happening, the "SIGNAL NEWNYM" event
        # will not be received. so we wait for it
        reactor.callLater(1, newid_no_signal, 10, all_done)
        yield all_done

cmd = NewidCommand()
__all__ = ['cmd']



from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log, failure
from twisted.internet import defer, reactor, error
import humanize

import txtorcon

from carml.interface import ICarmlCommand
from carml.util import dump_circuits, format_net_location, nice_router_name, colors, wrap

LOG_LEVELS = ["DEBUG", "INFO", "NOTICE", "WARN", "ERR"]


class MonitorOptions(usage.Options):
    optFlags = [
        ('once', 'o', 'Exit after printing the current state.'),
        ('no-streams', 's', 'Without this, list Tor streams.'),
        ('no-circuits', 'c', 'Without this, list Tor circuits.'),
        ('no-addr', 'a', 'Without this, list address mappings (and expirations, with -f).'),
        ('no-guards', 'g', 'Without this, Information about your current Guards.'),
        ('verbose', 'v', 'Additional information. Circuits: ip, location, asn, country-code.'),
    ]

    optParameters = [
        ('log-level', 'l', 'INFO', 'Subscribe to Tor log messages up to "level". Valid levels are: %s' % ' '.join(LOG_LEVELS)),
    ]

    def opt_log_level(self, x):
        if x.lower() == 'all':
            self['log-level'] = LOG_LEVELS

        levels = x.upper()
        if ',' in x:
            levels = levels.split(',')
        else:
            levels = [levels]

        for sub in levels:
            if sub not in LOG_LEVELS:
                raise RuntimeError('Unknown log level "%s".' % sub)
        self['log-level'].extend(levels)

    def __init__(self):
        """
        We override this to get rid of the Twisted default --version and --help things
        """
        super(MonitorOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self['log-level'] = []


def string_for_circuit(state, circuit):
    # path = '->'.join(map(lambda x: x.location.countrycode or '??', circuit.path))
    path = '->'.join(map(lambda x: x.name, circuit.path))
    state = circuit.state
    if state.lower() == 'failed':
        state = colors.red(state)
    return 'Circuit %d (%s) is %s for purpose "%s"' % (circuit.id, path, state, circuit.purpose)


def string_for_stream(state, stream):
    circ = ''
    if stream.circuit:
        circ = ' via circuit %d' % stream.circuit.id
    proc = txtorcon.util.process_from_address(stream.source_addr, stream.source_port, state)
    if proc:
        proc = ' from process "%s"' % (colors.bold(os.path.realpath('/proc/%d/exe' % proc)), )

    elif stream.source_addr == '(Tor_internal)':
        proc = ' for Tor internal use'

    elif stream.source_addr:
        proc = ' from remote "%s:%s"' % (str(stream.source_addr), str(stream.source_port))

    else:
        proc = ''
    hoststring = colors.bold('%s:%d' % (stream.target_host, stream.target_port))
    return 'Stream %d to %s %s%s%s' % (stream.id, hoststring, colors.green('attached'), circ, proc)


class StreamLogger(txtorcon.StreamListenerMixin):
    def __init__(self, state, verbose):
        self.state = state
        self.verbose = verbose

    def stream_attach(self, stream, circuit):
        print(string_for_stream(self.state, stream))
        if self.verbose:
            m = "  " + '->'.join(map(lambda x: nice_router_name(x), circuit.path))
            m += ' (%s)' % ' '.join(map(lambda r: str(r.location.countrycode), circuit.path))
            print(m)

    def stream_failed(self, stream, remote_reason='', **kw):
        print('Stream %d %s because "%s"' % (stream.id, colors.red('failed'),
                                             colors.red(remote_reason)))


def flags(d):
    """
    Basically converts a dict into name=value string, but only if the
    key is all-uppercase. This is because of the way we re-map keyword
    args for the circuit listeners, etcetra.
    """

    r = ''
    for (k, v) in d.iteritems():
        if k.upper() == k:
            r += '%s=%s ' % (k, v)
    return r


class CircuitLogger(txtorcon.CircuitListenerMixin):
    def __init__(self, state, show_flags=False):
        self.state = state
        self.show_flags = show_flags

    def circuit_launched(self, circuit):
        print(string_for_circuit(self.state, circuit))

    def circuit_extend(self, circuit, router):
        print(string_for_circuit(self.state, circuit))

    def circuit_built(self, circuit):
        print(string_for_circuit(self.state, circuit))
        if self.show_flags:
            flagslist = ['%s=%s' % x for x in circuit.flags.items()]
            flags = wrap(' '.join(flagslist), 72, '    ')
            print(colors.cyan('    Flags:'), flags.lstrip())

    def circuit_failed(self, circuit, **kw):
        print('Circuit %d failed (%s).' % (circuit.id, flags(kw)))

    def circuit_closed(self, circuit, **kw):
        print('Circuit %d %s lasted %s (%s).' % (circuit.id,
                                                 colors.red('closed'),
                                                 humanize.time.naturaldelta(circuit.age()),
                                                 flags(kw)))


class AddressLogger(object):
    zope.interface.implements(txtorcon.interface.IAddrListener)

    def addrmap_added(self, addr):
        print('New address mapping: "%s" -> "%s".' % (addr.name, addr.ip))

    def addrmap_expired(self, name):
        print('Address mapping for "%s" expired.' % name)


def tor_log(level, msg):
    print('%s: %s' % (level, msg))


def monitor_callback(options, state):
    follow_string = None
    if options['log-level'] and not options['once']:
        follow_string = 'Logging ('
        for event in options['log-level']:  # LOG_LEVELS:
            state.protocol.add_event_listener(event, functools.partial(tor_log, event))
            follow_string += event + ', '
            if event == options['log-level']:
                break
        follow_string = follow_string[:-2] + ')'
    if not options['no-streams']:
        if follow_string:
            follow_string += ' and Stream'
        else:
            follow_string = 'Stream'
        if len(state.streams):
            print("Current streams:")
            for stream in state.streams.values():
                print('  ' + string_for_stream(state, stream))
        else:
            print("No streams.")
        state.add_stream_listener(StreamLogger(state, options['verbose']))

    if not options['no-circuits']:
        if follow_string:
            follow_string += ' and Circuit'
        else:
            follow_string = 'Circuit'

        if len(state.circuits):
            print("Current circuits:")
            dump_circuits(state, verbose=options['verbose'])
        else:
            print("No circuits.")
        state.add_circuit_listener(CircuitLogger(state, show_flags=options['verbose']))

    if not options['no-guards']:
        if len(state.entry_guards):
            print("Current Entry Guards:")
            for (name, router) in state.entry_guards.iteritems():
                if not router.from_consensus:
                    if router.name:
                        print("  %s: %s (not in consensus)" % (router.name, router.id_hex))
                    else:
                        print("  %s (not in consensus)" % router.id_hex)

                else:
                    print(" ", router.id_hex, router.name, format_net_location(router.location))

        else:
            print("No Guard nodes!")

    if not options['no-addr']:
        if follow_string:
            follow_string += ' and Address'
        else:
            follow_string = 'Address'

        if len(state.addrmap.addr):
            print("Current address mappings:")
            for addr in state.addrmap.addr.values():
                print('  %s -> %s' % (addr.name, addr.ip))
            state.addrmap.add_listener(AddressLogger())
        else:
            print("No address mappings.")

    all_done = defer.Deferred()
    if not options['once']:
        print('')
        print("Following new %s activity:" % follow_string)

        def stop_reactor(arg):
            print("Tor disconnected.")
            all_done.callback(None)
        def error(fail):
            print(colors.red('Error:'), fail.getErrorMessage())
        state.protocol.on_disconnect.addErrback(error).addBoth(stop_reactor)

    else:
        all_done.callback(None)
    return all_done


class MonitorCommand(object):
    zope.interface.implements(ICarmlCommand)

    name = 'monitor'
    help_text = """General information about a running Tor; streams, circuits, address-maps and event monitoring."""
    controller_connection = True
    build_state = True
    options_class = MonitorOptions

    def validate(self, options, mainoptions):
        return
        at_least_one = False
        for k in ['streams', 'circuits', 'addr']:
            if options[k]:
                at_least_one = True
                break
        if not at_least_one and not (options['guards'] or options['log-level']):
            raise RuntimeError("Must specify at least one of --streams, --circuits, "
                               "--guards or --addr or --log-level")
        if not options['once'] and not (at_least_one or options['log-level']):
            raise RuntimeError("Can't follow just guards updates; add --once or exclude "
                               " fewer things.")
        if options['log-level'] and options['once']:
            raise RuntimeError("--log-level with --once doesn't make sense.")

    def run(self, options, mainoptions, state):
        return monitor_callback(options, state)


cmd = MonitorCommand()
__all__ = ['cmd']



from __future__ import print_function
import sys
import time
import functools

from zope.interface import implementer
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import defer
from twisted.web.iweb import IAgentEndpointFactory
from twisted.web.client import Agent
from twisted.internet import reactor
from twisted.internet.protocol import Protocol


from carml.interface import ICarmlCommand
from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


class CopyBinOptions(usage.Options):
    """
    """

    optFlags = [
    ]

    optParameters = [
        ('service', 's', None, 'The endpoint you were given to download from.'),
    ]

@implementer(IAgentEndpointFactory)
class EndpointFactory(object):
    def __init__(self, ep):
        self.endpoint = ep

    def endpointForURI(self, uri):
        '''IAgentEndpointFactory API'''
        return self.endpoint


def receive(req):
    d = defer.Deferred()
    req.deliverBody(BodyReceiver(d))
    return d


class BodyReceiver(Protocol):
    def __init__(self, d):
        self.data = ''
        self.done = d

    def dataReceived(self, data):
        self.data += data

    def connectionLost(self, reason):
        self.done.callback(self.data)


@implementer(ICarmlCommand, IPlugin)
class CopyBinCommand(object):
    """
    The opposite of "carml pastebin" -- download something from a
    pastebin, possibly using stealth tor authentication as well.
    """

    name = 'copybin'
    help_text = """Download something from a "pastebin" hidden-service."""
    build_state = False
    controller_connection = True
    options_class = CopyBinOptions

    def validate(self, options, mainoptions):
        "ICarmlCommand API"
        if not options['service']:
            raise RuntimeError("--service option required")

    @defer.inlineCallbacks
    def run(self, options, mainoptions, proto):
        "ICarmlCommand API"

        config = txtorcon.TorConfig(proto)
        yield config.post_bootstrap

        socks = yield proto.get_info('net/listeners/socks')
        socks = socks['net/listeners/socks']
        socks_host, socks_port = socks.split(':')

        args = options['service'].split(':')
        onion = args[1]
        cookie = args[2].split('=')[1]
        ep = txtorcon.TorClientEndpoint(
            onion, 80,
            socks_hostname=socks_host, socks_port=int(socks_port)
        )

        auth = '%s %s' % (onion, cookie)
        if auth not in config.HidServAuth:
            config.HidServAuth.append(auth)
        yield config.save()

        agent = Agent.usingEndpointFactory(reactor, EndpointFactory(ep))
        res = yield agent.request('GET', 'http://%s/' % onion)
        print("Response:", res.code)
        print("bytes:", res.length)
        data = yield receive(res)
        print(data)


# the IPlugin/getPlugin stuff from Twisted picks up any object from
# here than implements ICarmlCommand -- so we need to instantiate one
cmd = CopyBinCommand()
